/* This file is part of Mailfromd.
   Copyright (C) 2005, 2006, 2007, 2008, 2009, 2010 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <unistd.h>
#include <stdlib.h>
#include <sysexits.h>
#include "c-ctype.h"

#include <mailutils/mailutils.h>
#include <mailutils/libcfg.h>
#include <mailutils/libargp.h>
#include <mailutils/daemon.h>

#include "libmf.h"
#include "mfdb.h"
#include "mf-dbm.h"

char *state_dir = DEFAULT_STATE_DIR;
char *file_option;               /* File name for DB management commands */
struct db_format *format_option;
time_t expire_interval = 0;      /* When set, overrides the expire interval
				    for the selected database */
int all_option;                  /* Process all databases */
int predict_next_option; 
double predict_rate;             /* Prediction rate for --list --format=rates*/
void (*run)(int argc, char **argv) = NULL;


void
assert_db_format()
{
	if (!format_option) {
		mu_error(_("operation is not applicable"));
		exit(EX_USAGE);
	}
}

char *
get_db_name()
{
	assert_db_format();
	if (file_option)
		return file_option;
	if (format_option->dbname)
		return format_option->dbname;
	mu_error(_("database file name is not given"));
	exit(EX_USAGE);
}

int
convert_rate(const char *arg, double *rate)
{
        double count;
        char *p;

        count = strtod(arg, &p);
        if (*p) {
                time_t div;
                const char *endp;

                while (*p && c_isspace(*p))
                        p++;
                if (c_isalpha(*p)) {
                        if (strlen(p) > 3
                            && memcmp(p, "per", 3) == 0 && c_isspace(p[3]))
                                p += 3;
                } else if (c_ispunct(*p))
                        p++;

                if (parse_time_interval(p, &div, &endp)) {
                        mu_error(_("unrecognized time format (near `%s')"),
                                 endp);
                        return 1;
                }
                count /= div;
        }
        *rate = count;
        return 0;
}
                

void
mfdbtool_delete(int argc, char **argv)
{
	int i;
	char *name = get_db_name();

	for (i = 0; i < argc; i++)
		db_delete(name, argv[i]);
}

void
mfdbtool_list(int argc, char **argv)
{
	int rc;
	char *name = get_db_name();
	
	if (!format_option->print_item) {
		/* Should not happen */
		mu_error(_("list is not applicable to this DB format"));
		exit(EX_USAGE);
	}		

	rc = 0;
	if (argc) {
		int i;
		for (i = 0; i < argc; i++) 
			rc += db_list_item(name, argv[i],
					   format_option->print_item);
	} else
		rc = db_list(name, format_option->print_item);
	exit(rc != 0);
}

static int
db_proc_enumerator(struct db_format *fmt, void *data)
{
	int (*func)(char *name, db_expire_t exp) = data;
	
	if (fmt->expire) 
		func(fmt->dbname, fmt->expire);
	return 0;
}

void
mfdbtool_expire(int argc, char **argv)
{
	//priv_setup();FIXME
	if (all_option) 
		db_format_enumerate(db_proc_enumerator, db_expire);
	else {
		if (!format_option->expire) {
			mu_error(_("expire is not applicable to this DB format"));
			exit(EX_USAGE);
		}		
		exit(db_expire(get_db_name(), format_option->expire) != 0);
	}
}	

void
mfdbtool_compact(int argc, char **argv)
{
	//priv_setup();FIXME
	if (all_option) 
		db_format_enumerate(db_proc_enumerator, db_compact);
	else {
		if (!format_option->expire) {
			mu_error(_("compact is not applicable to this DB format"));
			exit(EX_USAGE);
		}
		exit(db_compact(get_db_name(), format_option->expire) != 0);
	}
}



const char *program_version = "mfdbtool (" PACKAGE_STRING ")";
static char doc[] = N_("mfdbtool -- Mailfromd database management tool");
static char args_doc[] = "";

enum {
	OPTION_STATE_DIRECTORY = 256,
	OPTION_DELETE,
	OPTION_LIST,
	OPTION_EXPIRE,
	OPTION_COMPACT,
	OPTION_IGNORE_FAILED_READS,
	OPTION_PREDICT_NEXT,
	OPTION_LOCK_RETRY_COUNT,
	OPTION_LOCK_RETRY_TIMEOUT,
	OPTION_ALL,
	OPTION_TIME_FORMAT
};

static struct argp_option options[] = {
#define GRP 10
	{ NULL, 0, NULL, 0,
	  N_("Database management commands"), GRP },
	{ "delete", OPTION_DELETE, NULL, 0,
	  N_("delete given entries from the database"), GRP+1 },
	{ "list", OPTION_LIST, NULL, 0,
	  N_("list given database (default cache)"), GRP+1 },
	{ "expire", OPTION_EXPIRE, NULL, 0,
	  N_("delete expired entries from the database"), GRP+1 },
	{ "compact", OPTION_COMPACT, NULL, 0,
	  N_("compact the database"), GRP+1 },
#undef GRP
#define GRP 20
	{ NULL, 0, NULL, 0,
	  N_("Command modifiers"), GRP },
	{ "file", 'f', N_("FILE"), 0,
	  N_("DB file name to operate upon"),
	  GRP+1 },
	{ "format", 'H', N_("FORMAT"), 0,
	  N_("format of the DB file to operate upon"), GRP+1 },
	{ "ignore-failed-reads", OPTION_IGNORE_FAILED_READS, NULL, 0,
	  N_("ignore failed reads while compacting the database"), GRP+1 },
	{ "predict", OPTION_PREDICT_NEXT, N_("RATE"), 0,
	  N_("predict when the user will be able to send next message"),
	  GRP+1 },
	{ "all", OPTION_ALL, NULL, 0,
	  N_("with --compact or --expire: apply the operation to all "
	     "available databases"), GRP+1 },
	{ "time-format", OPTION_TIME_FORMAT, N_("FMT"), 0,
	  N_("output timestamps using given format (default \"%c\")"), GRP+1 },
	{ "debug", 'd', N_("LEVEL"), 0,
	  N_("set debugging level"), GRP+1 },
	
#undef GRP
#define GRP 30
	{ NULL, 0, NULL, 0,
	  N_("Configuration modifiers"), GRP },
#if 0
	{ "lock-retry-count", OPTION_LOCK_RETRY_COUNT, N_("NUMBER"), 0,
	  N_("set maximum number of attempts to acquire the lock"), GRP+1 },
	{ "lock-retry-timeout", OPTION_LOCK_RETRY_TIMEOUT, N_("TIME"), 0,
	  N_("set timeout for acquiring the lockfile"), GRP+1 },
#endif
	{ "state-directory", OPTION_STATE_DIRECTORY, N_("DIR"), 0,
	  N_("set new program state directory"), GRP+1 },
	{ "expire-interval", 'e', N_("NUMBER"), 0,
	  N_("set database expiration interval to NUMBER seconds"), GRP+1 },
#undef GRP
	{ NULL }
};

static int
db_proc_set_expire(struct db_format *fmt, void *data)
{
	fmt->expire_interval = expire_interval;
	return 0;
}

static error_t
parse_opt(int key, char *arg, struct argp_state *state)
{
	switch (key) {
	case OPTION_STATE_DIRECTORY:
		mf_optcache_set_option("state-directory", arg);
		break;
#if 0
		/* FIXME */
	case OPTION_LOCK_RETRY_COUNT:
		mf_optcacheset_option("lock-retry-count", arg);
	case OPTION_LOCK_RETRY_TIMEOUT:
		mf_optcacheset_option("lock-retry-timeout", arg);
#endif
	case 'd':
		mf_optcache_set_option("debug", arg);
		break;
		
	case 'e': {
		time_t interval;
		const char *endp;
		if (parse_time_interval(arg, &interval, &endp)) {
			argp_error(state,
				   _("unrecognized time format (near `%s')"),
				   endp);
		}
		expire_interval = interval;
		break;
	}
		
	case OPTION_PREDICT_NEXT:
		format_option = db_format_lookup("rate");
		if (convert_rate(arg, &predict_rate) == 0)
			predict_next_option = 1;
		break;
		
	case OPTION_IGNORE_FAILED_READS:
		ignore_failed_reads_option = 1;
		break;

	case OPTION_ALL:
		all_option = 1;
		break;
		
	case OPTION_TIME_FORMAT:
		ignore_failed_reads_option = 1;
		break;
		
	case 'f':
		file_option = arg;
		break;

	case 'H':
		format_option = db_format_lookup(arg);
		if (format_option == NULL)
			argp_error(state, _("unknown database format: %s"),
				   arg);
		break;
		
	case OPTION_DELETE:
		run = mfdbtool_delete;
		break;
		
	case OPTION_LIST:
		run = mfdbtool_list;
		break;
		
	case OPTION_EXPIRE:
		run = mfdbtool_expire;
		break;
		
	case OPTION_COMPACT:
		run = mfdbtool_compact;
		break;
		
	case ARGP_KEY_FINI:
		if (!run)
			argp_error(state, _("you must specify one of --list, --delete, --expire or --compat"));
		if (all_option &&
		    !(run == mfdbtool_compact || run == mfdbtool_expire))
			argp_error(state,
				   _("--all is meaningful only with "
				     "--expire or --compact option"));
		if (all_option && format_option)
			argp_error(state,
				   _("--format is incompatible with --all"));
		if (!format_option)
			format_option = db_format_lookup("cache");
		if (!format_option) {
			mu_error("INTERNAL ERROR: cache database not defined");
			exit(1);
		}
		if (expire_interval)
			db_format_enumerate(db_proc_set_expire, NULL);
		break;
		
	default:
		return ARGP_ERR_UNKNOWN;
	}
	return 0;
}
	

static int
cb_debug(void *data, mu_config_value_t *arg)
{
	if (mu_cfg_assert_value_type(arg, MU_CFG_STRING))
		return 1;
	mu_debug_parse_spec(arg->v.string);
	return 0;
}

struct mu_cfg_param callout_cfg_param[] = {
	{ "database", mu_cfg_section, NULL, 0, NULL, NULL },
	{ "state-directory", mu_cfg_string, &state_dir, 0, NULL,
	  N_("Set program state directory."),
	  N_("dir") },
	{ "debug", mu_cfg_callback, NULL, 0, cb_debug,
	  N_("Set Mailfromd debug verbosity level.  Argument is a comma-"
             "separated list of debug specifications, each of which has the "
	     "following form:\n"
	     "  <module: string>[=<level: number>]\n"
	     "where <module> is the name of a Mailfromd module, and <level> "
	     "is the desired verbosity level for that module."),
	  N_("spec: list") },
	{ NULL }
};

static void
set_state_directory(void *value)
{
	state_dir = value;
}

static void
set_debug(void *value)
{
	mu_debug_parse_spec(value);
}

static struct mf_option_cache option_cache[] = {
	{ "state-directory", NULL, mf_option_string, set_state_directory },
	{ "debug", NULL, mf_option_string, set_debug },
	{ NULL }
};

static const char *capa[] = {
	"common",
	"debug",
	NULL
};

static struct argp argp = {
	options,
	parse_opt,
	args_doc,
	doc,
	NULL,
	NULL,
	NULL
};

static void
version(FILE *stream, struct argp_state *state)
{
	mailfromd_version("mfdbtool", stream);
}

void
xalloc_die()
{
	mu_error(_("not enough memory"));
        abort();
}

int
main(int argc, char **argv)
{
	int rc, index;

	mf_init_nls();
	if (!program_invocation_short_name)
		program_invocation_short_name = argv[0];
	argp_program_version_hook = version;

	libdbm_init();
	db_format_setup();
	database_cfg_init();

	mf_optcache_add(option_cache, 0, MF_OCF_NULL|MF_OCF_STATIC);

	mu_argp_init(program_version, "<" PACKAGE_BUGREPORT ">");
	mu_app_rcfile = SYSCONFDIR "/mfdbtool.conf";//FIXME
	rc = mfd_app_init(&argp, capa, callout_cfg_param, argc, argv,
			  0, &index, NULL, NULL);
	if (rc)
		exit(EX_CONFIG);

	mf_optcache_flush();

	argv += index;
	argc -= index;

	if (chdir(state_dir)) {
		mu_error(_("cannot change to %s: %s"),
			 state_dir, strerror(errno));
		exit(EX_OSERR);
	}

	run(argc, argv);
	
	exit(0);
}
	
	

